/** 
 * OSC server for MLD output. Supports both request and continuous modes.
 * by Andres Colubri. August 17 2007.
 * 
 */

/*
  License (based on zlib/libpng):

  Copyright (c) 2007 Andres Colubri.

  This software is provided 'as-is', without any express or implied warranty.
  In no event will the authors be held liable for any damages arising from
  the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely.
*/

import processing.serial.*;
import oscP5.*;
import netP5.*;

// Server parameters.
OscP5 oscServer;
int port = 5204;
String ip = "192.168.0.111";
NetAddress clientLocation;
OscMessage oscMessage;

// Packet parameters.
int maxMLDMsgs = 10;                           // Number of MLD messages send in a single OSC packet.
float sendTimeInterval = 100;                  // Minimum waitng time (in milisecs) between two consecutive OSC packets.
int msgSize = 42;                              // Size in bytes of an MLD message.

// Serial parameters.
Serial serialPort;                             // The serial port
byte[] serialInArray = new byte[msgSize];      // Where we'll put what we receive
int serialCount = 0;                           // A count of how many bytes we receive
boolean isReading = false;                     // Whether we're currently reading from the MLD box.
boolean useDelay = false;                      // Use delay when no messages are read from MLD box. Delay time is set in then nex var.
int waitDelay = 30;                            // Delay to wait for new incoming signal from the MLD box (in milisecs).
boolean sendRawMsg = false;                    // If true, the OSC packets are an exact copy of the serial messages sent by the MLD box.
boolean requestMode = false;                   // In the request mode, the server has to ask the MLD machine for more messages.
                                               // Otherwise, the MLD machine sends out messages continously. This mode has to be
                                               // configured in the Serial setup of the MLD software.

// Filter parameters: to set which MLD messages are not send over OSC.
boolean discardLostTrackers = true;
boolean discardDisabledTrackers = true;
boolean discardClippedTrackers = false;
boolean discard2DTrackers = false;

// Debug parameters.
boolean printMLDMsg = false;
boolean printSendInfo = true;

// Timer parameters.
boolean setFramerate = false;                  // Tries, if true, to fix the framerate to the value specified below.
int fixedFPS = 24;
int timeStart, timeNow;
int frameCounter;
float fps;

// Tracker parameters.
int numTrackers = 4;
int tracker = 0;
char[] trackerID = { 'a', 'b', 'c', 'd' };

// Data arrays and some counters needed for synchronization.
int lastOutIdx = 0;
int oscOutArraySize = msgSize * maxMLDMsgs;
byte[] oscOutArray = new byte[oscOutArraySize];
byte[] emptyOutArray = new byte[oscOutArraySize];
boolean outArrayIsFull = false;
float lastSendTime = 0;

void setup() 
{
    if (setFramerate) frameRate(fixedFPS);
    size(100, 100);
    fill(255);

    // Font setup. 
    PFont font;
    font = loadFont("ArialMT-48.vlw"); 
    textFont(font, 10);
  
    // Print a list of the serial ports, for debugging purposes:
    println(Serial.list());

    // I know that the first port in the serial list on my mac
    // is always my  Keyspan adaptor, so I open Serial.list()[0].
    // On Windows machines, this generally opens COM1.
    // Open whatever port is the one you're using.
    serialPort = new Serial(this, Serial.list()[0], 115200);

    oscServer = new OscP5(this, port);
    clientLocation = new NetAddress(ip, port);
    oscMessage = new OscMessage("/MLD");
        
    tracker = 0;
    if (requestMode)
        // Send an 'a' to start the MLD sending data for tracker 1.
        serialPort.write(trackerID[tracker]);

    // Initialize the empty out array...
    for (int i = 0; i < oscOutArraySize; i++) emptyOutArray[i] = 0;
      
    frameCounter = 0;
}

void draw() 
{
    background(0);
    text("Server running", 10, 30); 
  
    // Get any new serial data
    while (serialPort.available() > 0) serialEvent();

    if (useDelay)
    {
        // If there's no serial data, send again until we get some.
        // (in case you tend to start Processing before you start your
        // external device):
        if (isReading == false)
        {
            delay(waitDelay);
            if (requestMode) serialPort.write(trackerID[tracker]);
        }
    }

    printFPS();
}

void serialEvent() 
{
    processByte((byte)serialPort.read());
}

void processByte(byte inByte) 
{
    // Add the latest byte from the serial port to array:
    isReading = true;
    if (0 <= serialCount) serialInArray[serialCount] = inByte;
    serialCount++;
    
    // If we have 'msgSize' bytes:
    if (msgSize <= serialCount)
    {
        if ((char)serialInArray[41] != 'E')
        {
            println("Wrong end character! Correcting...");
            // The end of signal character is not in the right place. Scanning for it
            // in order to fix message reading.
            int i, nE = 0;
            for (i = 0; i < 41; i++)            
                if ((char)serialInArray[i] == 'E') 
                {
                    nE = i;
                    break;
                }
            // Serial count has a negative initial value to account for the shift needed to put the first
            // message character inside the first element of the array.
            serialCount = - nE - 1;
            println("...set a shift of " + (nE + 1));
            return;  
        }
      
        int n = CharToInt((char)serialInArray[1]);
        if (n != tracker + 1) tracker = n - 1;
           
        char trackerStatus = (char)serialInArray[3];
        boolean discard = (discardLostTrackers && (trackerStatus == 'L')) ||
                          (discardDisabledTrackers && (trackerStatus == 'D')) ||
                          (discardClippedTrackers && (trackerStatus == 'C')) ||
                          (discard2DTrackers && (trackerStatus == '2'));
        
        if (!discard)
        {
            // Sending out OSC packet with tracker info.
            if (printMLDMsg)
            {
                String s = new String(serialInArray);
                println(s);
            }
            
            // 'maxMLDMsgs' MLD messages are appended in the oscOutArraySize array, to avoid packet flooding.
            // Also, data is sent over OSC with a minimum time interval set in sendTimeInterval.
            outArrayIsFull = oscOutArraySize <= lastOutIdx;
            if (!outArrayIsFull)
            {
                System.arraycopy(serialInArray, 0, oscOutArray, lastOutIdx, msgSize);
                lastOutIdx = lastOutIdx + msgSize;
            }
  
            float time = millis();
            if (outArrayIsFull || (sendTimeInterval < time - lastSendTime))
            {
                oscMessage.clear();
                if (sendRawMsg)
                    // Sending raw OSC message.
                    oscMessage.add(oscOutArray);
                else
                {
                    // Separating raw data into integer variables.                  
                    int offset, num, msgCount = serialCount / msgSize;
                    for (int i = 0; i < msgCount; i++)
                    {
                        offset = msgSize * i;

                        oscMessage.add(CharToInt((char)oscOutArray[offset + 1]));
                        oscMessage.add((char)oscOutArray[offset + 3]);
                        oscMessage.add(CharArrayToInt(offset + 5, offset + 9));
                        oscMessage.add(CharArrayToInt(offset + 11, offset + 15));
                        oscMessage.add(CharArrayToInt(offset + 17, offset + 21));
                        oscMessage.add(CharArrayToInt(offset + 23, offset + 27));
                        oscMessage.add(CharArrayToInt(offset + 29, offset + 33));
                        oscMessage.add(CharArrayToInt(offset + 35, offset + 39));
                    }
                }
              
                lastSendTime = time;
                if (printSendInfo) println("Sending " + lastOutIdx + " bytes...");
                oscServer.send(oscMessage, clientLocation);
                
                System.arraycopy(emptyOutArray, 0, oscOutArray, 0, oscOutArraySize);
                lastOutIdx = 0;
                
                if (outArrayIsFull)
                {
                    System.arraycopy(serialInArray, 0, oscOutArray, lastOutIdx, msgSize);
                    lastOutIdx = lastOutIdx + msgSize;
                    outArrayIsFull = false;
                }
            }
        }
        
        if (tracker == numTrackers - 1) tracker = 0;
        else tracker = tracker + 1;
                
        if (requestMode)
            // Request new MLD readings for tracker:
            serialPort.write(trackerID[tracker]);
            
        // Reset serialCount:
        serialCount = 0;
        isReading = false;
    }
}

int CharToInt(char c)
{
    switch(c) 
    {
        case '-': return -1;
        case '0': return 0;
        case '1': return 1;
        case '2': return 2;
        case '3': return 3;
        case '4': return 4;
        case '5': return 5;
        case '6': return 6;
        case '7': return 7;
        case '8': return 8;
        case '9': return 9;
        default: return 0;
    }
}

int CharArrayToInt(int n1, int n2)
{
    int d, num;
    num = 0;
    for (int n = 0; n < n2 - n1 + 1; n++) 
    {
        if (oscOutArray[n] == ' ') break;
        d = CharToInt((char)oscOutArray[n2 - n]);     
        if (d == -1) num = -1 * num;
        else num = num + d * (int)pow(10, n);
    }
    return num;
}

void printFPS()
{
    timeNow = millis();
    frameCounter = frameCounter + 1;
    int diff = timeNow - timeStart;
    if (0 < diff) fps = 1000 * (float)frameCounter / (float)diff;
    text("FPS: " + fps, 10, 50); 
    if (timeNow % 1000 == 0)
    {
        timeStart = timeNow;
        frameCounter = 0;
    }
}
